import re
import os
import pytest
import signal
import psutil

import nb2workflow.service


@pytest.fixture
def test_notebook():
    return os.environ.get('TEST_NOTEBOOK',
                          os.path.join(os.getcwd(), "tests/testrepo/workflow-notebook.ipynb"))

@pytest.fixture
def test_notebook_old():
    return os.environ.get('TEST_NOTEBOOK',
                          os.path.join(os.getcwd(), "tests/testrepo/workflow-notebook.ipynb"))

@pytest.fixture
def test_notebook_repo():
    return os.environ.get('TEST_NOTEBOOK_REPO',
                          os.path.join(os.getcwd(), 'tests/testrepo/'))


@pytest.fixture
def app(test_notebook):
    app = nb2workflow.service.app
    app.notebook_adapters = nb2workflow.nbadapter.find_notebooks(test_notebook)
    nb2workflow.service.setup_routes(app)
    print("creating app")
    return app


def kill_child_processes(parent_pid, sig=signal.SIGTERM):
    try:
        parent = psutil.Process(parent_pid)
        children = parent.children(recursive=True)
        for process in children:
            process.send_signal(sig)
    except psutil.NoSuchProcess:
        return


@pytest.fixture
def service_fixture(pytestconfig, test_notebook_repo):
    import subprocess
    import os
    import copy
    import time
    from threading import Thread

    env = copy.deepcopy(dict(os.environ))
    print("rootdir", str(pytestconfig.rootdir))
    env['PYTHONPATH'] = str(pytestconfig.rootdir)+":" + \
        str(pytestconfig.rootdir)+"/tests:"+env.get('PYTHONPATH', "")
    print("pythonpath", env['PYTHONPATH'])

    p = subprocess.Popen(
        ["python", "-m", "nb2workflow.service",
            test_notebook_repo, '--port', '9292'],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=False,
        env=env,
    )

    url_store = [None]

    def follow_output():
        url_store[0] = None
        for line in iter(p.stdout.readline, b''):
            line = line.decode()

            print("following server:", line.rstrip())
            m = re.search("Running on (.*?) \(Press CTRL\+C to quit\)", line)
            if m:
                # alaternatively get from configenv
                url_store[0] = m.group(1)[:-1]
                print("found url:", url_store[0])

            if re.search("\* Debugger PIN:.*?", line):
                url_store[0] = url_store[0].replace("0.0.0.0", "127.0.0.1")
                print("server ready, url", url_store[0])

    thread = Thread(target=follow_output, args=())
    thread.start()

    while url_store[0] is None:
        time.sleep(0.1)
    time.sleep(0.5)

    ddservice = url_store[0]

    yield ddservice

    print("child:", p.pid)
    # p.kill()

    kill_child_processes(p.pid, signal.SIGKILL)
    os.kill(p.pid, signal.SIGKILL)


def pytest_addoption(parser):
    parser.addoption(
        "--runservice", action="store_true", default=False, help="run service tests"
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "service: mark test as relying on local service to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runservice"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_service = pytest.mark.skip(reason="need --runservice option to run")
    for item in items:
        if "service" in item.keywords:
            item.add_marker(skip_service)



import time
import atexit

from apscheduler.schedulers.background import BackgroundScheduler


def schedule_callable(f, interval):
    scheduler = BackgroundScheduler()
    scheduler.add_job(func=f, trigger="interval", seconds=interval)
    scheduler.start()

    atexit.register(lambda: scheduler.shutdown())



import argparse
import logging
import os

import cwlgen

import nb2workflow.nbadapter as nbadapter


def python_type2cwl_type(pt):
    if pt == str:
        return 'string'
    elif pt == bool:
        return 'boolean'

    return pt.__name__


def nb2cwl_container(image, notebook_fn, cwl_fn, command=None, nbrunner_module=None):
    nba = nbadapter.NotebookAdapter(notebook_fn)

    if command is None:
        base_command = "python"
        if nbrunner_module is None:
            arguments = ["-m", "nb2workflow.nbadapter", "/repo/"+nba.name]
        else:
            arguments = ["-m", nbrunner_module, "/repo/"+nba.name]
    else:
        base_command = 'bash'
        arguments = ['-c', command]

    tool_object = cwlgen.CommandLineTool(
        tool_id=nba.name,
        base_command=base_command,
        label=None,
        doc=None,
        cwl_version="v1.0",
        stdin=None,
        stderr=None,
        stdout=None,
        path=None)

    notebook_name = os.path.basename(notebook_fn)

    tool_object.arguments = arguments

    tool_object.requirements.append(
        cwlgen.DockerRequirement(docker_pull=image))

    for par in nba.extract_parameters().values():
        tool_object.inputs.append(
            cwlgen.CommandInputParameter(
                par['name'],
                param_type=python_type2cwl_type(par['python_type']),
                label=None,
                secondary_files=None,
                param_format=None,
                streamable=None,
                doc=None,
                input_binding=dict(prefix="--inp-"+par['name'], separate=True),
                default=par['default_value'],
            )
        )

    # tool_object.outputs.append(
    #    cwlgen.CommandOutputParameter('log',
    #                                  param_type='stdout',
    #                                  doc='log')
    # )

    for n, o in nba.extract_output_declarations().items():
        tool_object.outputs.append(
            cwlgen.CommandOutputParameter(n,
                                          param_type='string',
                                          doc='lines found with the pattern')
        )

    tool_object.export()
    tool_object.export(cwl_fn)


def nb2cwl(notebook_fn, cwl_fn, nbrunner_module="nb2workflow.nbadapter"):
    nba = nbadapter.NotebookAdapter(notebook_fn)

    tool_object = cwlgen.CommandLineTool(
        tool_id=nba.name,
        base_command="python",
        label=None,
        doc=None,
        cwl_version="v1.0",
        stdin=None,
        stderr=None,
        stdout=None,
        path=None)

    tool_object.arguments = ["-m", nbrunner_module, notebook_fn]

    for par in nba.extract_parameters().values():
        tool_object.inputs.append(
            cwlgen.CommandInputParameter(
                par['name'],
                param_type=python_type2cwl_type(par['python_type']),
                label=None,
                secondary_files=None,
                param_format=None,
                streamable=None,
                doc=None,
                input_binding=dict(
                    prefix="--inp-"+par['name']+"=", separate=False),
                default=par['default_value'],
            )
        )

    # tool_object.outputs.append(
    #    cwlgen.CommandOutputParameter('log',
    #                                  param_type='stdout',
    #                                  doc='log')
    # )

    for n, o in nba.extract_output_declarations().items():
        tool_object.outputs.append(
            cwlgen.CommandOutputParameter(n,
                                          param_type='string',
                                          doc='lines found with the pattern')
        )

    tool_object.export()
    tool_object.export(cwl_fn)


def main():
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('cwl', metavar='cwl', type=str)
    parser.add_argument('--publish', metavar='upstream-url',
                        type=str, default=None)
    parser.add_argument(
        '--publish-as', metavar='published url', type=str, default=None)
    parser.add_argument('--debug', action="store_true")
    parser.add_argument('--container', metavar="container")
    parser.add_argument(
        '--nbrunner-module', metavar="nbrunner_module", default="nb2workflow.nbadapter")

    args = parser.parse_args()

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    root = logging.getLogger()

    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    if args.debug:
        root.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

    if args.container:
        nb2cwl_container(args.container, os.path.basename(
            args.notebook), args.cwl, nbrunner_module=args.nbrunner_module)
    else:
        nb2cwl(args.notebook, args.cwl, nbrunner_module=args.nbrunner_module)


if __name__ == "__main__":
    main()



import argparse
import json
import os
import pathlib
import subprocess
import tempfile
import time
import ruamel.yaml as yaml
from . import version

default_config = {
    "config_schema_version": "0.1.0",
    "notebook_patterns": ["notebooks/*", "*.ipynb"],
    "extra_data": [],
    "use_repo_base_image": False
}


def determine_origin(repo):
    if os.path.isdir(repo):
        return subprocess.check_output(
            ["git", "remote", "get-url", "origin"],
            cwd=repo).decode().strip()
    else:
        return repo

def deploy(git_origin, deployment_base_name, namespace="oda-staging", local=False):
    git_origin = determine_origin(git_origin)

    with tempfile.TemporaryDirectory() as tmpdir:        
        subprocess.check_call(# cli is more stable than python API
            ["git", "clone", git_origin, "nb-repo"],
            cwd=tmpdir)

        local_repo_path = pathlib.Path(tmpdir) / "nb-repo"

        descr = subprocess.check_output( # cli is more stable than python API
            ["git", "describe", "--always", "--tags"],
            cwd=local_repo_path ).decode().strip()

        config_fn = local_repo_path / "mmoda.yaml"

        config = default_config.copy()
        if os.path.exists(config_fn):
            config.update(yaml.load(open(config_fn)))

        if not config['use_repo_base_image']: 
            open(pathlib.Path(tmpdir) / "Dockerfile", "a").write(f"""
FROM python:3.9

ADD nb-repo/requirements.txt /requirements.txt
RUN pip install -r requirements.txt

""")

        # we could use completely new image too. but lets keep renku etc in it
        open(pathlib.Path(tmpdir) / "Dockerfile", "a").write(f"""
RUN pip install nb2workflow[cwl,service,rdf]=={version()}

COPY nb-repo/ /repo/

RUN curl -o /usr/bin/jq -L https://github.com/stedolan/jq/releases/download/jq-1.5/jq-linux64; chmod +x /usr/bin/jq
RUN for nn in /repo/*.ipynb; do mv $nn $nn-tmp;  jq '.metadata.kernelspec.name |= "python3"' $nn-tmp > $nn ; rm $nn-tmp ; done

ENTRYPOINT nb2service --debug /repo/ --host 0.0.0.0 --port 8000 | cut -c1-500
""")

        image = f"odahub/nb-{pathlib.Path(git_origin).name}:{descr}-nb2w{version()}" # {time.strftime(r'%y%m%d%H%M%S')}"

        subprocess.check_call( # cli is more stable than python API
            ["docker", "build", ".", "-t", image],
            cwd=tmpdir)

        if local:
            subprocess.check_call( # cli is more stable than python API
                ["docker", "run", image],
                cwd=tmpdir)
        else:
            subprocess.check_call( # cli is more stable than python API
                ["docker", "push", image],
                cwd=tmpdir)


            deployment_name = deployment_base_name + "-backend"
            try:
                subprocess.check_call(
                    ["kubectl", "patch", "deployment", deployment_name, "-n", namespace,
                    "--type", "merge",
                    "-p", 
                    json.dumps(
                        {"spec":{"template":{"spec":{
                            "containers":[
                                {"name": deployment_name, "image": image}
                            ]}}}})
                    ]
                )
            except Exception as e:
                subprocess.check_call(
                    ["kubectl", "create", "deployment", deployment_name, "-n", namespace, "--image=" + image]
                )
                subprocess.check_call(
                    ["kubectl", "expose", "deployment", deployment_name, "--name", deployment_name, 
                    "--port", "8000", "-n", namespace]
                )
            


def main():
    parser = argparse.ArgumentParser(description='')
    parser.add_argument('repository', metavar='repository', type=str)
    parser.add_argument('deployment_name', metavar='deployment_name', type=str)
    parser.add_argument('--namespace', metavar='namespace', type=str, default="oda-staging")
    parser.add_argument('--local', action="store_true", default=False)
    
    args = parser.parse_args()
    
    deploy(args.repository, args.deployment_name, namespace=args.namespace, local=args.local)


if __name__ == "__main__":
    main()



from __future__ import print_function

import os
import sys
import argparse
import json
import docker
import shutil
import tempfile
import checksumdir
import subprocess

from nb2workflow import version
from nb2workflow.cwl import nb2cwl_container
from nb2workflow.nbadapter import find_notebooks

import logging

logger = logging.getLogger(__name__)


def build_python(dockerfile):
    dockerfile.append("RUN yum install -y python")
    dockerfile.append("RUN curl https://bootstrap.pypa.io/get-pip.py | python")


def import_repo(repo_source, target):
    print("importing repo", repo_source, "to", target)
    if os.path.isdir(repo_source):
        shutil.copytree(repo_source, target)
    else:
        raise NotImplementedError

    return checksumdir.dirhash(target)


def prepare_image(repo_source, from_image, service=True, nb2w_path=None, runprefix="", entrypoint=None):

    tempdir = tempfile.mkdtemp()

    rel_repo_path = "repo"
    repo_path = os.path.join(tempdir, rel_repo_path)

    repo_hash = import_repo(repo_source, repo_path)

    dockerfile = []

    dockerfile.append("FROM {}".format(from_image))
    dockerfile.append("ARG REPO_PATH=./{}".format(rel_repo_path))

    pipconf = os.path.join(repo_path, 'pip.conf')
    logger.info("using pipconf %s", pipconf)
    if os.path.exists(pipconf):
        dockerfile.append("ENV PIP_CONFIG_FILE=/etc/pip/pip.conf")
        dockerfile.append("ADD $REPO_PATH/pip.conf /etc/pip/pip.conf")

    dockerfile.append("ADD $REPO_PATH/requirements.txt /requirements.txt")
    dockerfile.append("RUN {} pip install --upgrade pip".format(runprefix))
    # ,repo_hash))
    dockerfile.append(
        "RUN {} pip install -r /requirements.txt --upgrade".format(runprefix))
    dockerfile.append("ADD $REPO_PATH /repo")
    dockerfile.append(
        "RUN {} touch /repo-hash-{}; pip install -r /repo/requirements.txt --upgrade".format(runprefix, repo_hash))
    dockerfile.append("RUN useradd -ms /bin/bash oda")

    if nb2w_path is None:
        dockerfile.append(
            "RUN {} pip install nb2workflow=={}".format(runprefix, version()))
    else:
        subprocess.check_output(
            ["git", "clone", nb2w_path, os.path.join(tempdir, "nb2workflow")])
        dockerfile.append("ADD ./nb2workflow /nb2workflow")
        dockerfile.append(
            "RUN {} cd /nb2workflow; pip install .".format(runprefix))

    dockerfile.append("USER oda")
    dockerfile.append("WORKDIR /workdir")

    if service:
        logger.info('service mode')
        dockerfile.append("ENTRYPOINT nb2service /repo/ --host 0.0.0.0")
    else:
        logger.info('not service mode')
        dockerfile.append('ENTRYPOINT []')
        #dockerfile.append('ENTRYPOINT ["bash"]' )
        #dockerfile.append('ENTRYPOINT ["nbrun", "/repo/"]' )

    open(os.path.join(tempdir, "Dockerfile"), "w").write(
        ("\n".join(dockerfile))+"\n")

    return tempdir


def build_image(tempdir, tag_image, nb2workflow_revision):
    cli = docker.from_env()

    print("-- building image, tagging as", tag_image)
    r = cli.api.build(
        path=tempdir,
        tag=tag_image,
        quiet=False,
        buildargs=dict(nb2workflow_revision=nb2workflow_revision),
        # stream=True,
        rm=True,
    )
    for k in r:
        try:
            print(json.loads(k)['stream'].strip())
        except:
            print(k)

    return True


def main():

    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('repo', metavar='repo', type=str)
    parser.add_argument('--run', action='store_true')
    parser.add_argument('--build', action='store_true')
    parser.add_argument('--job', action='store_true')
    parser.add_argument('--name', metavar='TAG', type=str, default="nb2worker")
    parser.add_argument('--from-image', metavar='FROM IMAGE',
                        type=str, default="python:3.6")
    parser.add_argument('--tag-image', metavar='TAG', type=str, default="")
    parser.add_argument('--host', metavar='host',
                        type=str, default="127.0.0.1")
    parser.add_argument('--port', metavar='port', type=int, default=9191)
    parser.add_argument('--nb2wrev', metavar='TAG', type=str, default="master")
    parser.add_argument('--nb2wpath', metavar='PATH', type=str)
    parser.add_argument('--volume', metavar='mount:mount', type=str, nargs="*", default=[])
    parser.add_argument('--docker-run-prefix',  type=str, default="")
    parser.add_argument('--store-dockerfile',
                        metavar='location', type=str, default=None)
    parser.add_argument('--docker-command', type=str, default=None)
    parser.add_argument('--entrypoint', type=str, default=None)

    args = parser.parse_args()

    repo_path = args.repo
    tag_image = args.tag_image

    if tag_image.lower() != tag_image:
        print("\033[31mdocker does not accept tags with anything but lowercase!\033[0m")
        print(f"\033[31mthis is bad: {tag_image}\033[0m")
        sys.exit(1)

    if args.tag_image == "":
        tag_image = os.path.basename(os.path.abspath(repo_path))

    tempdir = prepare_image(
        repo_path, args.from_image,
        service=not args.job,
        nb2w_path=args.nb2wpath,
        runprefix=args.docker_run_prefix,
        entrypoint=args.entrypoint,
    )

    if args.store_dockerfile:
        shutil.copy(os.path.join(tempdir, "Dockerfile"), args.store_dockerfile)
        print("\033[31mstored Dockerfile as\033[0m", args.store_dockerfile)

    if args.build:
        build_result = build_image(tempdir, tag_image, args.nb2wrev)

        if build_result is None:
            raise Exception("failed to build")

        print("built:", build_result)

        if args.job:

            for n, nba in find_notebooks(repo_path).items():
                nb2cwl_container(tag_image, nba.notebook_fn,
                                 n+".cwl", command=args.docker_command)

        if args.run:
            if not args.job:
                print("running", tag_image, "service on", args.port)
                cli = docker.from_env()
                c = cli.containers.run(
                    tag_image,
                    user=os.getuid(),
                    ports={9191: (args.host, args.port)},
                    name=args.name,
                    detach=True,
                    volumes=dict([
                        (os.getcwd(), {"bind": "/workdir", "mode": "rw"}),
                    ]+[v.split(":", 1) for v in args.volume]),
                )

                for r in c.attach(stream=True):
                    print(c, r.strip())
            else:
                print("running", tag_image)
                cli = docker.from_env()
                c = cli.containers.run(
                    tag_image,
                    user=os.getuid(),
                    name=args.name,
                    detach=True,
                    volumes=dict([
                        (os.getcwd(), {"bind": "/workdir", "mode": "rw"}),
                    ]+[v.split(":", 1) for v in args.volume]),
                    entrypoint=["nbrun", "/repo/*ipynb"],
                )

                for r in c.attach(stream=True):
                    print(c, r.strip())


if __name__ == "__main__":
    main()



import re
import logging
logger = logging.getLogger('nb2workflow.publish')

try:
    import consul
except:
    consul = None


def publish(upstream_url, name, service_host, service_port):
    if consul is None:
        return

    r = re.match(r"(.*?)://(.*?)(?:$|:)(\d*)",upstream_url)
    if r:
        scheme, host, port = r.groups()
    else:
        r = re.match(r"(.*?)(?:$|:)(\d*)",upstream_url)
        scheme = "http"
        host, port = r.groups()
    
    if port == "":
        port = 8500
    else:
        port = int(port)

    cc = consul.Consul(host = host, scheme = scheme, port = port)

    logger.debug("found services: %s",cc.agent.services())

    logger.debug("will publish as %s, %s",service_host, service_port)
    
    cc.agent.service.register(name, address = service_host, port = service_port, tags = ["nb2service", "traefik.protocol=https"])




import os
import subprocess

import pytest

import nb2workflow.cwl as cwl

@pytest.mark.xfail
def test_cwl(test_notebook):
    cwl_fn = "test.cwl"

    cwl.nb2cwl(test_notebook, cwl_fn)

    subprocess.check_call(["cwl-runner", cwl_fn])


@pytest.mark.xfail
def test_cwl_odakb(test_notebook):
    cwl_fn = "test.cwl"

    cwl.nb2cwl(test_notebook, "test.cwl", nbrunner_module="odakb.evaluator")

    subprocess.check_call(["cwl-runner", cwl_fn])



import json
import os
import requests
import time
import datetime
import logging
from collections import OrderedDict

from diskcache import Cache

from nb2workflow import nbadapter

cache = Cache('data/default-cache')
enable_cache = False

try:
    from nb2workflow import logstash
    logstasher = logstash.LogStasher()
except Exception as e:
    import logging
    logging.warning("unable to setup logstash %s",repr(e))

    logstasher = None

try:
    import sentry_sdk
    sentry_sdk.init(os.environ.get("SENTRY_URI", open("/cdci-resources/sentry-uri").read().strip()))
except ImportError:
    sentry_sdk = None
except Exception as e:
    import logging
    logging.debug("big problem with sentry: %s",repr(e))
    sentry_sdk = None


class WorkflowException(Exception):
    pass

def serialize_workflow_exception(e):
    try:
        return dict(
                    ename = e[0].ename,
                    evalue = e[0].evalue,
                    edump = e[1][0],
                )
    except TypeError:
        return dict(
                    ename = repr(e),
                    evalue = "",
                    edump = repr(e)
                )

    
def reroute(router, *args, **kwargs):
    workflow_routes = dict([ r.split("=") for r in os.environ.get('WORKFLOW_ROUTES','').split(",") if len(r.split("=")) == 2 ])

    workflow = args[0]

    if workflow in workflow_routes:
        r_w = workflow_routes[workflow]
        return r_w.split(":") + args, kwargs
    
    if workflow in os.environ.get('STAGING_WORKFLOWS','').split(','):
        return router+"-staging", args, kwargs

    return router, args, kwargs

def evaluate(router, *args, **kwargs):
    key = json.dumps((router, args, OrderedDict(sorted(kwargs.items()))))

    ntries = kwargs.pop('_ntries', 30)
    async_request = kwargs.pop('_async_request', True)
    cached = kwargs.pop('_cached', True)

    print("async_request is not used here, but is set to", async_request)


    if logstasher:
        logstasher.set_context(dict(router=router, args=args, kwargs=kwargs))
        logstasher.log(dict(event='starting'))

    if cached and enable_cache and key in cache:
        v = cache.get(key)
        print("restored from cache, key:", key)
        print("restored from cache, value:", v)

        if v == {} or v is None:
            print("this value is empty, regenerate")
        else:
            return v

    print("before routing", router, args, kwargs)
    router, args, kwargs = reroute(router, *args, **kwargs)
    print("after routing", router, args, kwargs)

    if router == "localfile":
        location = args[0]
        args = args[1:]

        nba = nbadapter.NotebookAdapter(location+"/%s.ipynb"%args[0])

        # unused args

        params = kwargs

        print("calling",params)

        exceptions = nba.execute(params,
                    log_output=True,
                    progress_bar=False)

        output = nba.extract_output()

        result = dict(output = output, exceptions = [serialize_workflow_exception(e) for e in exceptions])

    elif router.startswith("odahub") or router.startswith("host"):
        workflow = args[0]

        if router == "odahub-staging":
            url_template = f"https://oda-workflows-{workflow}-staging.odahub.io/api/v1.0/get/" + "{}"

        if router == "odahub":
            url_template = f"https://oda-workflows-{workflow}.odahub.io/api/v1.0/get/" + "{}"
    
        if router == "host":
            url_template = args[0]+"/api/v1.0/get/{}"

        url = url_template.format(*args[1:])
        print("url:",url)

        ntries = ntries
        while ntries > 0:
            try:
                print("towards",ntries,url,kwargs)
                c=requests.get(
                    url=url,
                    params=kwargs,
                    auth=requests.auth.HTTPBasicAuth("cdci", open("/cdci-resources/reproducible").read().strip())
                )
                print("decoding",c.text)

                try:
                    result = c.json()
                except Exception as ed:
                    print("problem decoding:", repr(ed))
                    print("raw output:",c.text)
                    logstasher.log(dict(event='failed to decode output',raw_output=c.text, exception=repr(ed)))
                    raise


                if 'output' in result and 'workflow_status' in result['output']:
                    if result['output']['workflow_status'] != "done": # bad
                        print("waiting for async workflow")
                        time.sleep(5)

                        ntries -= 1
                        continue

                break

            except Exception as e:
                print("problem from service", repr(e))

                logstasher.log(dict(event='problem evaluating',exception=repr(e)))
                
                if ntries <= 1:
                    if sentry_sdk:
                        sentry_sdk.capture_exception()
                    raise

                time.sleep(5)

                ntries -= 1

        if 'output' not in result:
            result = dict(output=result)

                #raise
    else:
        raise NotImplementedError


    if logstasher:
        logstasher.log(dict(event='done'))

    cache.set(key, result)
    print("stored to cache", key)

    return result




import socket
import json
import sys
import os

import collections

def flatten(d, parent_key='', sep='.'):
    items = []
    for k, v in d.items():
        #print("k,v",k,v)
        new_key = str(parent_key) + sep + str(k) if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        elif isinstance(v, collections.Iterable) and not isinstance(v,str):
            items.extend(flatten(dict(enumerate(v)), new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class LogStasher:
    def __init__(self, url=None):
        if url is None:
            self.url = os.environ.get("LOGSTASH_ENTRYPOINT", open("/cdci-resources/logstash-entrypoint").read().strip())
        else:
            self.url = url

        self.context = {}

    def set_context(self, c):
        self.context = c
    
    def log(self, msg):
        HOST, PORT = self.url.split(":")
        PORT = int(PORT)

        msg = flatten(dict(list(self.context.items()) + list(msg.items())))


        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except Exception as e:
            print("[ERROR] %s\n" % repr(e)) 
            

        try:
            sock.connect((HOST, PORT))
        except Exception as e:
            print("[ERROR] %s\n" % repr(e)) 

        sock.send(json.dumps(msg).encode())

        sock.close()



from ast import literal_eval
import hashlib
import os
import sys
import glob
import shutil
import yaml 
import re
import time
import tempfile
import pprint
import subprocess
import yaml
import argparse
import json
import base64

import papermill as pm
import scrapbook as sb
import nbformat
from nbconvert import HTMLExporter

from nb2workflow.health import current_health
from nb2workflow import workflows

import logging
logger=logging.getLogger(__name__)


try:
    from nb2workflow import logstash
    logstasher = logstash.LogStasher()
except Exception as e:
    logger.debug("unable to setup logstash %s",repr(e))

    logstasher = None


class PapermillWorkflowIncomplete(Exception):
    pass


def cast_parameter(x,par):
    logger.debug("cast %s %s",x,par)
    if par['python_type'] is bool:
        if x in ['false', 'False', 0, '0', '']:
            return False
        elif x in ['true', 'True', 1, '1']:
            return True
        else:
            raise ValueError(f'Parameter {par["name"]} value "{x}" can not be interpreted as boolean.')
    return par['python_type'](x)

def understand_comment_references(comment):
    logger.debug("treating comment %s",comment)

    oda_ontology_prefix = "http://odahub.io/ontology"
    r = re.search(r"\b("+oda_ontology_prefix+r".*?)(?:\s+|$)", comment)
    if r:
        owl_type = r.groups()[0]
        logger.debug("comment contains owl references: %s",owl_type)
    else:
        owl_type = None
        logger.debug("no references in this comment")

    return dict(
        owl_type = owl_type,
    )


def parse_nbline(line):
    if line.strip()=="":
        return None
    elif line.strip().startswith("#"):
        logger.debug("found detached comment: \"%s\"",line)
        return None
    else:
        if "#" in line:
            assignment_line,comment=line.split("#",1)
        else:
            assignment_line=line
            comment=""
            
        if "=" in assignment_line:
            name, value_str = assignment_line.split("=", 1)
            name = name.strip()
        else:
            name = assignment_line.strip()
            value_str=None

        try:
            value=literal_eval(value_str.strip())
            python_type = type(value)
        except Exception:
            value = value_str
            python_type = str

        comment=comment

        return dict(
                    name = name,
                    value = value,
                    python_type = python_type,
                    comment = comment,
                    owl_type = understand_comment_references(comment).get('owl_type',None),
                )


class InputParameter:
    def __init__(self):
        pass

    @classmethod
    def from_nbline(cls,line):
        r = parse_nbline(line)
        if r is None:
            return r
        else:
            obj = cls()
            obj.raw_line=line

            p = parse_nbline(line)
            
            obj.name = p['name']
            obj.default_value = p['value']
            obj.python_type = p['python_type']
            obj.comment = p['comment']
            obj.owl_type = p['owl_type']

            obj.choose_owl_type()
            
            logger.debug("%s %s %s comment: %s",obj.name,obj.default_value.__class__,obj.default_value,obj.comment)
            return obj
    

    def choose_owl_type(self):
        self.owl_type = None

        if self.comment.strip() != "":
            references =  understand_comment_references(self.comment)

            if references.get('owl_type',None):
                self.owl_type = references.get('owl_type')

        if self.owl_type is None:
            self.owl_type = "http://www.w3.org/2001/XMLSchema#"+self.python_type.__name__ # also use this if already defined

    def as_dict(self):
        return dict(
                    default_value=self.default_value,
                    python_type=self.python_type,
                    name=self.name,
                    comment=self.comment,
                    owl_type=self.owl_type,
                )


class NotebookAdapter:
    def __init__(self,notebook_fn):
        self.notebook_fn = notebook_fn
        self.name = notebook_short_name(notebook_fn)
        logger.debug("notebook adapter for %s",notebook_fn)
        logger.debug(self.extract_parameters())

    def new_tmpdir(self):
        logger.debug("tmpdir was "+getattr(self,'_tmpdir','unset'))
        self._tmpdir = None
        logger.debug("tmpdir became %s", self._tmpdir)

        return self.tmpdir

    @property
    def tmpdir(self):
        if getattr(self,'_tmpdir', None) is None:
            self._tmpdir = tempfile.mkdtemp(prefix="nb2w-")
        return self._tmpdir
    
    @property
    def preproc_notebook_fn(self):
        return os.path.join(self.tmpdir,os.path.basename(self.notebook_fn.replace(".ipynb","_preproc.ipynb")))

    @property
    def output_notebook_fn(self):
        return os.path.join(self.tmpdir,os.path.basename(self.notebook_fn.replace(".ipynb","_output.ipynb")))

    def read(self):
        return nbformat.reads(open(self.notebook_fn).read(), as_version=4)

    _notebook_origin = None

    @property
    def notebook_origin(self):
        if self._notebook_origin is None:
            notebook_dir = os.path.dirname(self.notebook_fn)
            logger.info('notebook_dir: %s', notebook_dir)

            url = subprocess.check_output(["git", "remote", "get-url", "origin"], cwd=notebook_dir).decode().strip()
            revision = subprocess.check_output(["git", "describe", "--always", "--tags"], cwd=notebook_dir).decode().strip()

            self._notebook_origin = f"{url}#{revision}"
        
        return self._notebook_origin

    @property
    def unique_name(self):
        return f"{self.name}_{hashlib.md5(self.notebook_origin.encode()).hexdigest()[:8]}"

    def export_html(self, fn=None):
        if fn is None:
            fn = "{}_output.html".format(self.name)

        if False:
            html_exporter = HTMLExporter()
            html_exporter.template_file = 'basic'
            (body, resources) = html_exporter.from_notebook_node(self.read())
            open(fn, "w").write(body)
        else:
            logging.info("converting... {}".format(subprocess.check_call(["jupyter", "nbconvert", "--to", "html", self.output_notebook_fn, '--output', os.path.abspath(fn)])))

        logger.info("exported html to %s", fn)

        return fn

    def extract_parameters(self):
        nb=self.read()

        input_parameters = {}
        system_parameters = {}

        for cell in nb.cells:
            if 'parameters' in cell.metadata.get('tags',[]):
                for line in cell['source'].split("\n"):
                    par=InputParameter.from_nbline(line)
                    if par is not None:
                        input_parameters[par.name]=par.as_dict()
                        input_parameters[par.name]['value']=par.as_dict()['default_value']
            
            if 'system-parameters' in cell.metadata.get('tags',[]):
                for line in cell['source'].split("\n"):
                    par=InputParameter.from_nbline(line)
                    if par is not None:
                        system_parameters[par.name]=par.as_dict()
            
            if 'injected-parameters' in cell.metadata.get('tags',[]):
                for line in cell['source'].split("\n"):
                    par=InputParameter.from_nbline(line)
                    if par is not None:
                        input_parameters[par.name]['value']=par.as_dict()['default_value']

        self.system_parameters = system_parameters

        return input_parameters
    
    def interpret_parameters(self,parameters):
        expected_parameters=self.extract_parameters()
        request_parameters=dict()

        unexpected_parameters=[]
        issues=[]

        for arg in parameters:
            if arg.startswith("_"): continue

            logger.info("request arg %s",parameters[arg])
            if arg in expected_parameters:
                try:
                    request_parameters[arg]=cast_parameter(parameters.get(arg),expected_parameters.get(arg))
                    logger.info("request arg %s provided as %s",parameters[arg],request_parameters[arg])
                except ValueError as e:
                    issues.append(e.args[0])
            else:
                unexpected_parameters.append(arg)

        if len(unexpected_parameters)>0:
            issues+=[f'found unexpected request parameters: {", ".join(unexpected_parameters)}, can be {", ".join(expected_parameters.keys())}']
            
        return dict(
                        issues=issues,
                        request_parameters=request_parameters,
                    )

    def update_summary(self, **d):
        if not hasattr(self, '_summary'):
            self._summary = dict(
                                name=self.name,
                                initialized=dict(s_epoch=time.time(), isot=time.strftime("%Y-%m-%d %H:%M:%S")),
                            )

        state=d.pop('state', None)

        self._summary.update(d)
        
        if state is not None:
            self._summary['state'] = self._summary.get("state",[]) + [(time.time(), state)]

        fn = os.path.join(self.tmpdir, "summary.yaml")
        yaml.dump(self._summary, open(fn, "w"))

        

    def execute(self, parameters, progress_bar = True, log_output = True, inplace=False):
        t0 = time.time()
        if logstasher is not None:
            logstasher.log(dict(origin="nb2workflow.execute", event="starting", parameters=parameters, workflow_name=notebook_short_name(self.notebook_fn), health=current_health()))

        logger.info("starting job")
        exceptions = self._execute(parameters, progress_bar, log_output, inplace)
            
        tspent = time.time() - t0
        if logstasher is not None:
            logstasher.log(dict(origin="nb2workflow.execute", 
                                event="done", 
                                parameters=parameters, 
                                workflow_name=notebook_short_name(self.notebook_fn), 
                                exceptions=list(map(workflows.serialize_workflow_exception, exceptions)),
                                health=current_health(), 
                                time_spent=tspent))

        return exceptions

    def _execute(self, parameters, progress_bar = True, log_output = True, inplace=False):

        if not inplace :
            tmpdir = self.new_tmpdir()
            logger.info("new tmpdir: %s", tmpdir)

            try:
                output = subprocess.check_output(["git","clone",os.path.dirname(os.path.realpath(self.notebook_fn)), tmpdir])
                logger.info("git clone output: %s", output)
            except Exception as e:
                logger.warning("git clone failed: %s, will attempt copytree", e)

                os.rmdir(tmpdir)

                shutil.copytree(os.path.dirname(os.path.realpath(self.notebook_fn)), tmpdir)
        else:
            tmpdir =os.path.dirname(os.path.realpath(self.notebook_fn))
            logger.info("executing inplace, no tmpdir is input dir: %s", tmpdir)

        
        self.update_summary(state="started", parameters=parameters)

        self.inject_output_gathering()
        exceptions = []

        
#        root = logging.getLogger()
#        root.setLevel(logging.DEBUG)

#        handler = logging.StreamHandler()
#        handler.setLevel(logging.DEBUG)
#        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
#        handler.setFormatter(formatter)
#        root.addHandler(handler)

#        root.info("towards excution")


        ntries = 10
        while ntries > 0:
            try:
                pm.execute_notebook(
                   self.preproc_notebook_fn,
                   self.output_notebook_fn,
                   parameters = parameters,
                   progress_bar = False,
                   log_output = True,
                   cwd = tmpdir, 
                )
            except pm.PapermillExecutionError as e:
                exceptions.append([e,e.args])
                logger.info(e)
                logger.info(e.args)
            
                if e.ename == "WorkflowIncomplete":
                    logger.info("detected incomplete workflow")
                    self.update_summary(state="incomplete dependency", dependency=repr(e))
                    raise  PapermillWorkflowIncomplete()

            except nbformat.reader.NotJSONError:
                ntries -= 1
                logger.info("retrying... %s", ntries)
                time.sleep(2)
                continue   

            break

        if len(exceptions) == 0:
            self.update_summary(state="done")
        else:
            self.update_summary(state="failed", exceptions=list(map(workflows.serialize_workflow_exception, exceptions)))

        return exceptions

    def extract_pm_output(self):
        nb = sb.read_notebook(self.output_notebook_fn)

        outputs=dict()
        for i, d in nb.scraps.dataframe.iterrows():
            logger.debug("d... %s",d)
            #if d.dtype == "record":
            outputs[d['name']]=d['data']

        return outputs

    
    def extract_output_declarations(self):
        nb=self.read()

        outputs = {}

        for cell in nb.cells:
            if 'outputs' in cell.metadata.get('tags',[]):
                for line in cell['source'].split("\n"):
                    p = parse_nbline(line)
                    if p is None: continue
                    outputs[p['name']] = p


        return outputs 

    def extract_output(self):
        return self.extract_pm_output()

    def inject_output_gathering(self):
        outputs = self.extract_output_declarations()

        output_gather_content="""
import papermill as pm
import scrapbook as sb
import base64
import json
import os
    
from nb2workflow.nbadapter import denumpyfy

"""
        for output in outputs.keys():
            logger.debug("output: %s",output)
            output_gather_content+="""
try:
    sb.glue("{output}",denumpyfy({output}))
except Exception as e:
    print("failed to glue {output}", {output})
    print("will glue jsonified")
    sb.glue("{output}",json.dumps(denumpyfy({output})))
""".format(output=output)

            output_gather_content+="\nisinstance({output},str) and os.path.exists({output}) and sb.glue(\"{output}_content\",base64.b64encode(open({output},'rb').read()).decode())".format(output=output)
            output_gather_content+="\n".format(output=output)

        nb = self.read()

        newcell = nbformat.v4.new_code_cell(source=output_gather_content)
        newcell.metadata['tags'] = ['injected-gather-outputs']

        nb = self.read()

        if nbformat.current_nbformat !=  nb.nbformat:
            logger.error("we assume nbformat version %s, but provided notebook is version %s, refusing!", 
                          nbformat.current_nbformat,
                          nb.nbformat)
            raise RuntimeError("incompatabile notebook major version")

        logger.info("provided notebook nbformat version minor %s while nbformat package minor version %s",
                    nb.nbformat_minor, nbformat.current_nbformat_minor)
                
        if nbformat.current_nbformat_minor == nb.nbformat_minor:
            logger.info("versions of notebook and environment match")
        elif nbformat.current_nbformat_minor < nb.nbformat_minor:
            logger.warn("notebook is newer than envionment package! please update your system or expect warnings")
        elif  nbformat.current_nbformat_minor > nb.nbformat_minor:
            logger.warning("will attempt to convert, but expect other warnings!")                            

            nb = nbformat.v4.convert.upgrade(nb, from_minor=nb.nbformat_minor)
        else:
            raise NotImplementedError


        nb.cells = nb.cells + [newcell] 

        logger.info("stored pre-processed notebook as %s", self.preproc_notebook_fn)
        pm.iorw.write_ipynb(nb, self.preproc_notebook_fn)

    def get_system_parameter_value(self, name, default):
        if name in self.system_parameters:
            return self.system_parameters.pop(name)['default_value'] 

        return default


def notebook_short_name(ipynb_fn):
    return os.path.basename(ipynb_fn).replace(".ipynb","")

def find_notebooks(source):

    if os.path.isdir(source):
        notebooks=[ fn for fn in glob.glob(source+"/*ipynb") if "output" not in fn and "preproc" not in fn ]
        logger.debug("found notebooks: %s",notebooks)

        if len(notebooks)==0:
            raise Exception("no notebooks found in the directory:",source)

        notebook_adapters=dict([
                (notebook_short_name(notebook),NotebookAdapter(notebook)) for notebook in notebooks
            ])
        logger.debug("notebook adapters: %s",notebook_adapters)


    elif os.path.isfile(source):
        notebook_adapters={notebook_short_name(source): NotebookAdapter(source)}

    else:
        raise Exception("requested notebook not found:",source)

    return notebook_adapters

def nbinspect(nb_source, out=True):
    nbas = find_notebooks(nb_source)

    class CustomEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, type):
                return str(obj)
            return json.JSONEncoder.default(self, obj)

    for n, nba in nbas.items():
        print(json.dumps(nba.extract_parameters(), indent=4, sort_keys=True, cls=CustomEncoder))


def nbreduce(nb_source, max_size_mb):
    cellsize_limit = None
    largest_cellsize = None


    while True:
        current_size_mb = os.path.getsize(nb_source)/1024./1024
        logging.info('notebook %s size %.4lg Mb', nb_source, current_size_mb)


        nb = nbformat.reads(open(nb_source).read(), as_version=4)

        newcells = []
        outputs_left = 0
        for i_cell, cell in enumerate(nb.cells):
            logging.info('try to reduce CELL #%i', i_cell)
    
            try:
                cellsize = len(json.dumps(cell.outputs))
            except AttributeError:
                logging.info("cell has no outputs, ignoring")
                continue

            if largest_cellsize is None or largest_cellsize < cellsize:
                largest_cellsize = cellsize

            logging.info("cell size %.5lg largest cell size %.5lg", cellsize, largest_cellsize)
            
            if 'injected-gather-outputs' in cell.metadata.get('tags', []):
                logging.info("is injected-gather-outputs: skipping")
                continue

            if cellsize_limit is not None and cellsize >= cellsize_limit:
                logging.info('cleaning cell')
                cell.outputs = []
                largest_cellsize = None


            if cell.outputs != []:
                logging.info('this cell has viable outputs')
                outputs_left += 1
        
            newcells.append(cell)

        if outputs_left == 0:
            logging.info('notebook size %.4lg Mb, and not more outputs left, cleaning aborted', current_size_mb)
            return
        else:
            logging.info('notebook size %.4lg Mb, still has %i outputs: cleaning may continue', current_size_mb,  outputs_left)

        nb.cells = newcells 
        pm.iorw.write_ipynb(nb, nb_source)

        if current_size_mb < max_size_mb:
            logging.info('notebook size %.4lg Mb is smaller than required %.5lg Mb, only cleaning gathering', current_size_mb, max_size_mb)
            return
        else:
            logging.info('notebook size %.4lg Mb is larger than required %.5lg Mb, setting cell size limit to the largest cell %s', 
                         current_size_mb, 
                         max_size_mb, 
                         cellsize_limit)

            cellsize_limit = largest_cellsize


def nbrun(nb_source, inp, inplace=False):

    nbas = find_notebooks(nb_source)

    if len(nbas) > 1:
        nba = nbas[inp.pop('notebook')]
    elif len(nbas) == 1:
        nba = list(nbas.values())[0]

    r = nba.interpret_parameters(inp)
    
    if r['issues'] != []:
        raise Exception(r['issues'])

    pars = r['request_parameters']

    logging.info("found parameters %s", repr(pars))

    exceptions = nba.execute(pars, inplace=inplace)

    if len(exceptions) == 0:
        logging.info("execution SUCCESSFUL!")
    else:
        logging.error("FAILED: %s", exceptions)

        with open("{}_exceptions.json".format(nba.name), "w") as f:
            json.dump(list(map(workflows.serialize_workflow_exception, exceptions)), f)

        fn = nba.export_html()
        open("{}_output.ipynb".format(nba.name), "wb").write(open(nba.output_notebook_fn, "rb").read())

        raise Exception(f"FAILED to execute {nba.name} {exceptions} html exported in {fn}")

    r={}
    for k,v in nba.extract_output().items():
        r[k.strip()]=repr(v)

    with open("{}_output.json".format(nba.name), "w") as f:
        json.dump(r, f)
    
    with open("cwl.output.json", "w") as f:
        json.dump(r, f)

    nbdata = open(nba.output_notebook_fn, "rb").read()
    nbfn = "{}_output.ipynb".format(nba.name)
    open(nbfn, "wb").write(nbdata)

    #TODO: store if not too big?   
    #r['output_notebook'] = nbfn
    #r['output_notebook_content'] = base64.b64encode(nbdata).decode()
        
    htmlfn = "{}_output.html".format(nba.name)
    nba.export_html(htmlfn)
    
    r['output_notebook_html'] = htmlfn
    r['output_notebook_html_content'] = base64.b64encode(open(htmlfn, "rb").read()).decode()

    return r


def traverse_structure(structure, modifier):
    if isinstance(structure, dict):
        return { k:traverse_structure(v, modifier) for k,v in structure.items() }

    if isinstance(structure, list):
        return [ traverse_structure(v, modifier) for v in structure ]

    return modifier(structure)



def denumpyfy(data):
    import numpy as np

    def numpy_adapter(d):
        if isinstance(d, np.bool_):
            return bool(d)

        if isinstance(d, np.ndarray):
            return d.tolist()
        
        if isinstance(d, np.float16) or \
           isinstance(d, np.float32) or \
           isinstance(d, np.float64):
            return float(d)

        if isinstance(d, np.int16) or \
           isinstance(d, np.int32) or \
           isinstance(d, np.int64):
            return int(d)

        return d

    return traverse_structure(
                data,
                numpy_adapter,
           )

def main_reduce():
    parser = argparse.ArgumentParser(description='Reduce notebook size') 
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('maxsizeMb', metavar='max_size_mb', type=float)
    parser.add_argument('--debug', action="store_true")
    
    args = parser.parse_args()

    setup_logging(args.debug)

    nbreduce(args.notebook, args.maxsizeMb)

def main_inspect():
    parser = argparse.ArgumentParser(description='Inspect some notebooks') # run locally, remotely, semantically
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('--debug', action="store_true")
    
    args = parser.parse_args()

    setup_logging(args.debug)

    nbinspect(args.notebook)

def setup_logging(debug=False):
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    root = logging.getLogger()
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    if debug:
        root.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

def main():
    parser = argparse.ArgumentParser(description='Run some notebooks') # run locally, remotely, semantically
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('--debug', action="store_true")
    parser.add_argument('--inplace', action="store_true")
    
    parser.add_argument('inputs', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    inputs={}
    for i in args.inputs:
        if not i.startswith("--inp-"): continue
        k,v = i.replace('--inp-','').split("=")
        inputs[k] = v
        
    setup_logging(args.debug)


    nbrun(args.notebook, inputs, inplace=args.inplace)


if __name__ == "__main__":
    main()



import io
import logging
import argparse
from typing import Optional

import nb2workflow.nbadapter as nbadapter

logger = logging.getLogger(__name__)

import requests
import rdflib

try:
    import odakb.sparql
except Exception:
    logger.error("some error")

try:
    import owlready2
        
    xsd = owlready2.get_ontology("https://www.w3.org/2001/XMLSchema#").load()
    kees = owlready2.get_ontology("http://linkeddata.center/kees/v1#").load()

    G = rdflib.Graph()
    open("function.xml", "w").write(G.load(io.StringIO(requests.get("https://raw.githubusercontent.com/FnOio/fnoio.github.io/master/ontology/0.4.1/function.ttl").text), format="turtle").serialize(format="xml"))

    fno = owlready2.get_ontology("function.xml").load()
    # fno.base_iri="https://w3id.org/function/ontology#"

    odaworkflow = owlready2.get_ontology("http://odahub.io/ontology/workflow#")
except Exception as e:
    logger.warning('unable to import owlready2: %s', e)
    owlready2=None

def get_dda():
    if owlready2 is None:
        return

    return owlready2.get_ontology("http://ddahub.io/ontology/analysis#")


def to_xsd_type(p):
    # if owlready2 is None:
    #     return

    out_type='string'

    if issubclass(p['python_type'],int):
        out_type='integer'

    if issubclass(p['python_type'],float):
        out_type='double'

    if issubclass(p['python_type'],str):
        out_type='string'

    logger.debug("owl type cast from %s to %s",p,repr(out_type))
    
    return p.get('owl_type', "http://www.w3.org/2001/XMLSchema#"+out_type)


#TODO: return owl option as an option

def function_semantic_signature(function_name, location, parameters, output, domains):
    G = rdflib.Graph()
    
    wfl = rdflib.URIRef(f'http://odahub.io/workflows#{function_name}')    
    
    oda_ns = rdflib.Namespace('http://odahub.io/ontology#')
    rdf_ns = rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')
    wfl_p_ns = rdflib.Namespace(f'http://odahub.io/workflows/{function_name}/parameter_bindings#')
    
    G.bind('oda', oda_ns)
    G.bind('rdfs', rdf_ns)
    # G.bind('wfl', wfl_p_ns)

    G.add((wfl, rdf_ns['type'], oda_ns['workflow']))
    G.add((wfl, oda_ns['location'], rdflib.Literal(location)))

    for pn, pv in parameters.items():
        logger.info('function_semantic_signature parameter pn=%s pv=%s', pn, pv)
        p_uri = wfl_p_ns[pn]
        logger.info('function_semantic_signature parameter p_uri=%s', p_uri)
        G.add((p_uri, rdf_ns['type'], rdflib.URIRef(to_xsd_type(pv))))
        G.add((wfl, oda_ns['expects'], p_uri))

    if domains is not None:
        for domain in domains:
            G.add((wfl, oda_ns['domain'], oda_ns[domain[0]]))
                    
    return G


def service_semantic_signature(nbas, format="xml", domains=None) -> str:
    G = rdflib.Graph()

    for target, nba in nbas.items():
        logger.info("target: %s nba: %s", target, nba)
        S_G = function_semantic_signature(
                                    nba.unique_name,
                                    location=nba.notebook_origin,
                                    parameters=nba.extract_parameters(),
                                    output=nba.extract_output_declarations(),
                                    domains=domains
                                    )

        for t in S_G:
            G.add(t)


    rdf_str = G.serialize(format=format)

    logger.debug(rdf_str)
    
    return rdf_str




def service_semantic_signature_owl(nbas, format="rdfxml"):
    if owlready2 is None:
        return
    dda = get_dda()
    dda.graph.destroy()
    dda = get_dda()

    with dda:
        class DataAnalysis(fno.Function):
            pass
        
        class WebDataAnalysis(DataAnalysis):
            pass
        
    r=[]
    for target, nba in nbas.items():
        r.append(function_semantic_signature(dda,target,
                                    parameters=nba.extract_parameters(),
                                    output=nba.extract_output_declarations()))

    f=io.BytesIO()
    dda.save(f, format=format)
    owl_str=f.getvalue().decode("unicode_escape")
    owl_str = bytes(owl_str, "utf-8").decode("unicode_escape")

    logger.debug(owl_str)
    
    return str(owl_str)




def nb2rdf(notebook_fn: str, domains: Optional[list]=None) -> str:
    nba = nbadapter.NotebookAdapter(notebook_fn)

    rdf = service_semantic_signature(dict(local=nba), format="turtle", domains=domains)
            
    logging.getLogger().info("rdf: %s", rdf)

    return rdf


def main():
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('--out-rdf', metavar='rdf', type=str)
    parser.add_argument('--domain', dest='domain', nargs="*", action='append')
    parser.add_argument('--publish', action="store_true")
    parser.add_argument('--debug', action="store_true")

    args = parser.parse_args()

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    root = logging.getLogger()

    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    if args.debug:
        root.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

    logger.error('domain %s', args)

    rdf = nb2rdf(args.notebook, domains=args.domain)

    if args.out_rdf:
        with open(args.out_rdf, "wt") as f:
            f.write(rdf)

    if args.publish:
        G = rdflib.Graph()
        G.parse(data=rdf)

        odakb.sparql.insert("\n".join([
            f"{s.n3()} {p.n3()} {o.n3()} ." for s, p, o in G
        ]))
    


if __name__ == "__main__":
    main()



from __future__ import print_function

import os
import json
import pytest
import base64

import nb2workflow.service
import nb2workflow.nbadapter
from flask import url_for




def test_service(client):
    r=client.get('/api/v1.0/options')
    
    service_name,service_signature=sorted(r.json.items())[0]
    print(service_signature)

    assert len(service_signature['parameters']) == 5

    #TODO: assert here paremeters

    print('get: /api/v1.0/get/'+service_name)

    r=client.get('/api/v1.0/get/'+service_name,query_string=dict(eminFAKE=20.))
    assert r.status_code == 400
    
    print(r.json['issues'])
    assert len(r.json['issues'])==1


    r=client.get('/api/v1.0/get/'+service_name,query_string=dict(emin=20.))
    assert r.status_code == 200

    print(r.json)

    open("output.png","wb").write(base64.b64decode(r.json['output']['spectrum_png_content']))

# trace
    r=client.get('/api/v1.0/options')
    
    service_name,service_signature=sorted(r.json.items())[0]

    print('get: /api/v1.0/get/'+service_name)

    r=client.get('/api/v1.0/get/'+service_name,query_string=dict(emin=20.))
    assert r.status_code == 200

    print(r.json)
    
    r=client.get('/trace/list')
    assert r.status_code == 200

    for l in sorted(r.json, key=lambda x:x['ctime']):
        print(l)

    job = r.json[-1]['fn'].split("/")[-1]
    
    r=client.get('/trace/'+job)

    print("job", r.json)

#    open("output.png","wb").write(base64.b64decode(r.json['output']['spectrum_png_content']))



def test_service_repo(client):
    
    r=client.get('/api/v1.0/options')
    
    service_signature=r.json['workflow-notebook']
    print(service_signature)

    assert len(service_signature['parameters']) == 5

    r=client.get('/api/v1.0/get/workflow-notebook',query_string=dict(eminFAKE=20.))
    assert r.status_code == 400
    
    print(r.json)
    print(r.json['issues'])
    assert len(r.json['issues'])==1


    r=client.get('/api/v1.0/get/workflow-notebook',query_string=dict(emin=20.))
    assert r.status_code == 200

    print(r.json)

    open("output.png","wb").write(base64.b64decode(r.json['output']['spectrum_png_content']))



def test_service_async_repo(client):
    
    r=client.get('/api/v1.0/options')
    
    service_signature=r.json['workflow-notebook']
    print(service_signature)

    assert len(service_signature['parameters']) == 5

    callback_fn = 'callback.json'
    
    r=client.get('/api/v1.0/get/workflow-notebook',
                 query_string=dict(
                     emin=20., 
                     _async_request='yes', 
                     _async_request_callback='file://' + callback_fn))

    assert r.status_code == 201

    print(r.json)

    from nb2workflow.service import AsyncWorker

    AsyncWorker('test-worker').run_one()

    callback_json = json.load(open(callback_fn))

    assert callback_json['action'] == 'done'

    r=client.get('/api/v1.0/get/workflow-notebook',
                query_string=dict(
                    emin=20., 
                    _async_request='yes', 
                    _async_request_callback='file://' + callback_fn))

    assert r.json['workflow_status'] == 'done'

    open("output.png","wb").write(base64.b64decode(r.json['data']['output']['spectrum_png_content']))



from __future__ import print_function
import re
from werkzeug.routing import RequestRedirect, MethodNotAllowed, NotFound
import queue
from nb2workflow import ontology, publish, schedule
from nb2workflow.nbadapter import NotebookAdapter, find_notebooks, PapermillWorkflowIncomplete

import os
import json
import glob
import time
import logging
import inspect
import requests
import base64
import hashlib
import datetime
import tempfile
import nbformat
import yaml

from io import BytesIO


from flask import Flask, make_response, jsonify, request, url_for, send_file, Response
from flask.json import JSONEncoder
from flask_caching import Cache
from flask_cors import CORS

from flasgger import LazyJSONEncoder, LazyString, Swagger, swag_from


from nb2workflow.workflows import serialize_workflow_exception

import threading

verify_tls = False


logger = logging.getLogger('nb2workflow.service')


async_queue = queue.Queue()


class ReverseProxied(object):
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get('HTTP_X_FORWARDED_PREFIX', '')
        if script_name:
            environ['SCRIPT_NAME'] = script_name
            path_info = environ['PATH_INFO']
            if path_info.startswith(script_name):
                environ['PATH_INFO'] = path_info[len(script_name):]

        scheme = environ.get('HTTP_X_SCHEME', '')
        if scheme:
            environ['wsgi.url_scheme'] = scheme
        return self.app(environ, start_response)


class CustomJSONEncoder(LazyJSONEncoder):
    def default(self, obj, *args, **kwargs):
        try:
            if isinstance(obj, type):
                return dict(type_object=repr(obj))
            iterable = iter(obj)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, obj)


cache = Cache(config={'CACHE_TYPE': 'simple'})


def create_app():
    app = Flask(__name__)

    template = {
        "swaggerUiPrefix": LazyString(lambda: request.environ.get('HTTP_X_FORWARDED_PREFIX', '')),
        "swagger": "2.0",
        "info": {
            "title": "ODAHub API",
            "description": "",
            "contact": {
                "responsibleOrganization": "ODA",
                "responsibleDeveloper": "Volodymyr SAVCHENKO",
                "email": "volodymyr.savchenko@unige.ch",
                "url": "https://odahub.io",
            },
            "termsOfService": "http://me.com/terms",
            "version": "0.0.1"
        }
    }
    swagger = Swagger(app, template=template)
    app.wsgi_app = ReverseProxied(app.wsgi_app)
    app.json_encoder = CustomJSONEncoder
    cache.init_app(app, config={'CACHE_TYPE': 'SimpleCache'})


#    CORS(app)
    return app


app = create_app()

app.async_workflows = dict()
app.started_at = datetime.datetime.now()


@app.after_request
def after_request(response):
    response.headers.add('Access-Control-Allow-Origin', '*')
    response.headers.add('Access-Control-Allow-Headers',
                         'Content-Type,Authorization')
    response.headers.add('Access-Control-Allow-Methods',
                         'GET,PUT,POST,DELETE,OPTIONS')
    return response

class AsyncWorker(threading.Thread):
    def __init__(self, worker_id):
        self.worker_id = worker_id
        super(AsyncWorker, self).__init__()

    def run(self):
        while True:
            self.run_one()
            time.sleep(5)

    def run_one(self):
        logger.info("worker_id %s", self.worker_id)
        async_workflow = async_queue.get(block=True)
        async_workflow.run()

class AsyncWorkflow:
    def __init__(self, key, target, params, callback=None):
        self.key = key
        self.target = target
        self.params = params
        self.callback = callback

        logger.info("initializing %s, callback", self, callback)

    def run(self):
        try:
            self._run()
        except Exception as e:
            logger.error("run failed unexplicably: %s", repr(e))
            app.async_workflows[self.key] = dict(
                output={}, 
                exceptions=[serialize_workflow_exception(e)]
            )

    def note(self, *args, **kwargs):
        if not hasattr(self, 'notes'):
            self.notes = []

        self.notes.append(dict(
            time=time.time(),
            data=(args, kwargs),
        ))

    blocked_until = 0

    def _run(self):
        if self.blocked_until > time.time():
            logger.info("workflow still blocked, waiting %i",
                        self.blocked_until - time.time())
            async_queue.put(self)
            app.async_workflows[self.key] = 'submitted'
            return

        app.async_workflows[self.key] = 'started'

        template_nba = app.notebook_adapters.get(self.target)

        nba = NotebookAdapter(template_nba.notebook_fn)

        try:
            exceptions = nba.execute(self.params['request_parameters'])
        except PapermillWorkflowIncomplete as e:
            logger.info("found incomplete workflow: %s, rescheduling", repr(e))

            self.note("rescheduled")

            self.blocked_until = time.time() + 10

            async_queue.put(self)
            app.async_workflows[self.key] = 'submitted'            

            return

        logger.info("exceptions: %s", repr(exceptions))

        if len(exceptions) > 0:
            output = 'incomplete'
            logger.error("exceptions: %s", repr(exceptions))
        else:
            nretry = 10
            while nretry > 0:
                try:
                    output = nba.extract_output()
                    logger.info("completed, output length %s", len(output))
                    if len(output) == 0:
                        logger.debug(
                            "output from notebook is empty, something failed, attempts left: %s", nretry)
                    else:
                        break
                except nbformat.reader.NotJSONError as e:
                    logger.debug(
                        "output notebook incomplete %s attempts left: %s", e, nretry)
                except Exception as e:
                    logger.debug(
                        "output notebook incomplte or does not exist %s attempts left: %s", e, nretry)

                nretry -= 1
                time.sleep(1)

        logger.debug("output: %s", output)

        logger.info("updating key %s", self.key)
        app.async_workflows[self.key] = dict(output=output, exceptions=list(
            map(serialize_workflow_exception, exceptions)), jobdir=nba.tmpdir)

        self.perform_callback()

    def perform_callback(self):
        if self.callback is None:
            logger.info('no callback registered, skipping')
            return

        logger.info('will perform callback: %s', self.callback)


        result = app.async_workflows[self.key]

        callback_payload = dict(
            action='done'
        )
        
        if re.match('^file://', self.callback):
            with open(self.callback.replace('file://', ''), "w") as f:
                 json.dump(callback_payload, f)
            logger.info('stored callback in a file %s', self.callback)

        elif re.match('^https?://', self.callback):
            r = requests.get(self.callback, params=callback_payload)
            logger.info('callback %s returns %s : %s', self.callback, r, r.text)
        
        else:
            raise NotImplementedError


def workflow(target, background=False, async_request=False):
    issues = []

    async_request = request.args.get('_async_request', async_request)
    async_request_callback = request.args.get('_async_request_callback', None)

    logger.debug("target %s", target)

    if not background:
        logger.debug("raw parameters %s", request.args)

    template_nba = app.notebook_adapters.get(target)
    nba = NotebookAdapter(template_nba.notebook_fn)

    if nba is None:
        interpreted_parameters = None
        issues.append("target not known: %s; available targets: %s" %
                      (target, app.notebook_adapters.keys()))
    else:
        if not background:
            interpreted_parameters = nba.interpret_parameters(request.args)
            issues += interpreted_parameters['issues']
        else:
            interpreted_parameters = dict(request_parameters=[])

    logger.debug("interpreted parameters %s", interpreted_parameters)

    # async
    if async_request:
        key = hashlib.sha224(json.dumps(
            dict(target=target, params=interpreted_parameters)).encode('utf-8')).hexdigest()

        value = app.async_workflows.get(key, None)

        print('cache key/value', key, value)

        if value is None:
            async_task = AsyncWorkflow(
                key=key, target=target, params=interpreted_parameters, callback=async_request_callback)

            async_queue.put(async_task)

            app.async_workflows[key] = 'submitted'
            return make_response(jsonify(workflow_status="submitted", comment="task created"), 201)

        elif value in ['started', 'submitted']:
            return make_response(jsonify(workflow_status=value, comment="task is "+value), 201)

        else:
            return make_response(jsonify(workflow_status="done", data=value, comment=""), 200)

    if len(issues) > 0:
        return make_response(jsonify(issues=issues), 400)
    else:
        exceptions = nba.execute(interpreted_parameters['request_parameters'])

        nretry = 10
        while nretry > 0:
            try:
                output = nba.extract_output()
                if len(output) == 0:
                    logger.debug(
                        "output from notebook is empty, something failed, attempts left: %s", nretry)
                else:
                    break
            except nbformat.reader.NotJSONError as e:
                logger.debug(
                    "output notebook incomplte %s attempts left: %s", e, nretry)

            nretry -= 1
            time.sleep(1)

        logger.debug("output: %s", output)
        logger.debug("exceptions: %s", exceptions)

        r = jsonify(dict(
                    output=output,
                    exceptions=[repr(e) for e in exceptions],
                    jobdir=nba.tmpdir,
                    ))

        return_code = 200
        if len(exceptions) > 0:
            return_code = 500

        return r, return_code


def to_oapi_type(in_type):
    if issubclass(in_type, bool):
        out_type = 'boolean'
    
    elif issubclass(in_type, int):
        out_type = 'integer'

    elif issubclass(in_type, float):
        out_type = 'number'

    elif issubclass(in_type, str):
        out_type = 'string'

    elif issubclass(in_type, bool):
        out_type = 'boolean'

    else:
        out_type = 'string'
        logger.debug(f"using default type cast from {in_type}to {out_type}")
    

    logger.debug("oapi type cast from %s to %s", repr(in_type), repr(out_type))

    return out_type


def get_view_function(url, method='GET'):
    """Match a url and return the view and arguments
    it will be called with, or None if there is no view.
    """

    adapter = app.url_map.bind('localhost')

    try:
        match = adapter.match(url, method=method)
    except RequestRedirect as e:
        # recursively match redirects
        return get_view_function(e.new_url, method)
    except (MethodNotAllowed, NotFound):
        # no match
        return None

    try:
        # return the view function and arguments
        return app.view_functions[match[0]], match[1]
    except KeyError:
        # no view is associated with the endpoint
        return None


def setup_routes(app):
    for target, nba in app.notebook_adapters.items():
        target_specs = {
            "parameters": [
                {
                    "name": p_name,
                    "in": "query",
                    "type": to_oapi_type(p_data['python_type']),
                    "required": False,
                    "default": p_data['default_value'],
                    "description": p_data['comment']+" "+p_data['owl_type'],
                }
                for p_name, p_data in nba.extract_parameters().items()
            ],
            "responses": {
                "200": {
                    "description": repr(nba.extract_output_declarations()),
                }
            }
        }

        endpoint = 'endpoint_'+target

        def funcg(target):
            def workflow_func():
                return workflow(target)
            return workflow_func

        logger.debug("target: %s with endpoint %s", target, endpoint)

        def response_filter(rv):
            if isinstance(rv, tuple) and isinstance(rv[0], Response) and rv[1] != 200:
                logger.info("NOT caching response %s", rv[1])
                return False
            elif isinstance(rv, Response) and rv.status != 200:
                logger.info("NOT caching response %s", rv)
                return False
            else:
                logger.info("caching response %s", rv)
                return True

        cache_timeout = nba.get_system_parameter_value('cache_timeout', 0)
        try:
            app.route('/api/v1.0/get/'+target, methods=['GET'], endpoint=endpoint)(
                swag_from(target_specs)(
                    cache.cached(timeout=cache_timeout, response_filter=response_filter, query_string=True)(
                        funcg(target)
                    )))
        except AssertionError as e:
            logger.warning("unable to add route: %s, ignoring the endpoint", e)
            continue

        schedule_interval = nba.get_system_parameter_value(
            'schedule_interval', 0)
        if schedule_interval > 0:
            logger.info("scheduling callable %s every %lg",
                        str(funcg), float(schedule_interval))

            def schedulable():
                with app.test_request_context('/api/v1.0/get/'+target):
                    from flask import request
                    get_view_function('/api/v1.0/get/'+target)[0]()

            schedule.schedule_callable(schedulable, schedule_interval)

# list input -> output function signatures and identities


@app.route('/api/v1.0/options', methods=['GET'])
def workflow_options():
    return jsonify(dict([
        (
                        target,
                        dict(output=nba.extract_output_declarations(),
                             parameters=nba.extract_parameters()),
                        )
        for target, nba in app.notebook_adapters.items()
    ]))


@app.route('/api/v1.0/get-<mode>/<target>/<filename>', methods=['GET'])
def workflow_filename(mode, target, filename):

    target_url = url_for('endpoint_'+target, _external=True, **request.args)

    # report equivalency

    if "HTTP_AUTH" in os.environ:
        username, password = os.environ.get("HTTP_AUTH").split(":")
        auth = requests.auth.HTTPBasicAuth(username, password)
    else:
        auth = None

    r = requests.get(target_url, verify=verify_tls, auth=auth)

    try:
        rj = r.json()
        logger.info(rj.keys())

        if 'data' in rj:
            output = rj.get('data').get('output')
        elif 'output' in rj:
            output = rj.get('output')
        else:
            return jsonify(rj)

        if filename+'_content' in output:
            content = base64.b64decode(output.get(filename+'_content', None))
        else:
            return jsonify({'workflow_status': 'anomaly', 'comment': 'searching for key '+filename+'_content'+', available: '+(", ".join(output.keys())), 'base_workflow_result': rj})

        if content:
            if mode == "file" or mode == "png":
                return send_file(BytesIO(content), mimetype='image/png')
            elif mode == "html":
                return content
            else:
                return 404
        else:
            return jsonify(dict(
                exceptions=["no such file, available:", output.keys()],
            ))

    except Exception as e:
        logger.error("problem decoding: %s", repr(e))
        return jsonify(dict(
            exceptions={"problem decoding": repr(
                e), "raw_response": r.content.decode('utf-8')},
        ))


@app.route('/api/v1.0/rdf', methods=['GET'])
def workflow_rdf():
    return make_response(app.service_semantic_signature)


@app.route('/health')
def healthcheck():
    status, issues = current_health()

    if len(issues) == 0:
        return jsonify(dict(summary="all is ok!", status=status))
    else:
        return make_response(jsonify(issues=issues, status=status, summary="warning: "+"; ".join(issues)), 500)


def current_health():
    issues = []
    status = {}

    statvfs = os.statvfs(".")
    status['fs_space'] = dict(
        size_mb=statvfs.f_frsize * statvfs.f_blocks / 1024 / 1024,
        avail_mb=statvfs.f_frsize * statvfs.f_bavail / 1024 / 1024,
    )

    if status['fs_space']['avail_mb'] < 300:
        issues.append("not enough free space: %.5lg Mb left" %
                      status['fs_space']['avail_mb'])

    import psutil

    processes = []
    status['n_open_files'] = 0
    status['n_processes'] = 0
    status['n_threads'] = 0

    for proc in psutil.process_iter():
        try:
            processes.append(dict(
                n_open_files=len(proc.open_files()),
            ))

            status['n_open_files'] += len(proc.open_files())
            status['n_processes'] += 1
            status['n_threads'] += proc.num_threads()
        except Exception as e:
            pass

    status['cpu_times'] = dict(psutil.cpu_times_percent()._asdict())

    status['loadavg'] = psutil.getloadavg()

    if max(status['loadavg']) > 10:
        issues.append("high load avg: %s" % repr(status['loadavg']))

    status['disk_usage'] = dict([(k+"_mb", v/1024/1024) if k != "percent" else (k, v)
                                 for k, v in dict(psutil.disk_usage(".")._asdict()).items()])

    status['async'] = dict(qsize=async_queue.qsize(),
                           async_workflows_n=len(app.async_workflows))

    #status['processes'] = processes

    return status, issues


@app.route('/test')
def test():
    results = {}
    expecting = []

    for template_nba in app.notebook_adapters.values():
        if template_nba.name.startswith('test_'):
            key = template_nba.name

            if key in app.async_workflows:
                print("found", app.async_workflows[key])

                if isinstance(app.async_workflows[key], dict):
                    print("found result seems a reasonable dict")
                    workflow_status = app.async_workflows[key].get(
                        'workflow_status', 'done')
                    print("workflow_status", workflow_status)
                else:
                    workflow_status = app.async_workflows[key]

                if workflow_status == 'done':
                    # and output notebook
                    results[template_nba.name] = app.async_workflows[key]['exceptions']
                    print("workflow_status is done, results exceptions:",
                          results[template_nba.name])
                else:
                    expecting.append(
                        dict(key=key, workflow_status=workflow_status))
            else:
                async_task = AsyncWorkflow(key=key, target=template_nba.name, params=dict(
                    request_parameters=dict(location=os.path.dirname(template_nba.notebook_fn))))

                async_queue.put(async_task)

                app.async_workflows[key] = 'started'
                expecting.append(dict(key=key, workflow_status='submitted'))

    if expecting != []:
        return make_response(jsonify(dict(expecting=expecting)), 201)
    else:
        if all([v in ['[]', []] for v in results.values()]):
            print("tests passed")
            return make_response('all is OK: '+"; ".join(results.keys()), 200)
        else:
            print("tests NOT passed")
            return make_response(jsonify(results), 500)


@app.route('/')
def root():
    issues = []

    if len(issues) == 0:
        return "all is ok!"
    else:
        return make_response(jsonify(issues=issues), 500)


def main():
    import argparse

    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('notebook', metavar='notebook', type=str)
    parser.add_argument('--host', metavar='host',
                        type=str, default="127.0.0.1")
    parser.add_argument('--port', metavar='port', type=int, default=9191)
    parser.add_argument('--async-workers', metavar='N', type=int, default=3)
    #parser.add_argument('--tmpdir', metavar='tmpdir', type=str, default=None)
    parser.add_argument('--publish', metavar='upstream-url',
                        type=str, default=None)
    parser.add_argument(
        '--publish-as', metavar='published url', type=str, default=None)
    parser.add_argument(
        '--profile', metavar='service profile', type=str, default="oda")
    parser.add_argument('--debug', action="store_true")
    parser.add_argument('--one-shot', metavar='workflow', type=str)

    args = parser.parse_args()

    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    root = logging.getLogger()

    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    root.addHandler(handler)

    if args.debug:
        root.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

    app.notebook_adapters = find_notebooks(args.notebook)
    setup_routes(app)
    app.service_semantic_signature = ontology.service_semantic_signature(
        app.notebook_adapters)

    if args.publish:
        logger.info("publishing to %s", args.publish)

        if args.publish_as:
            s = args.publish_as.split(":")
            publish_host, publish_port = ":".join(s[:-1]), int(s[-1])
        else:
            publish_host, publish_port = args.host, args.port

        for nba_name, nba in app.notebook_adapters.items():
            publish.publish(args.publish, nba_name, publish_host, publish_port)

  #  for rule in app.url_map.iter_rules():
 #       logger.debug("==>> %s %s %s %s",rule,rule.endpoint,rule.__class__,rule.__dict__)

    for worker_i in range(args.async_workers):
        async_worker = AsyncWorker('default-%i' % worker_i)
        async_worker.start()

    app.run(host=args.host, port=args.port)


@app.route('/status')
def status():
    return jsonify(
        version=os.environ.get('WORKFLOW_VERSION', 'unknown'),
        started_at=app.started_at.strftime("%s"),
        started_since=(datetime.datetime.now()-app.started_at).seconds,
        background_jobs=len([w for w in app.async_workflows if w]),
        stored_jobs=len(app.async_workflows),
    )


@app.route('/async/delete')
def async_delete():
    return jsonify(app.async_workflows)


@app.route('/async/clear')
def async_clear():
    s = app.async_workflows
    app.async_workflows = dict()
    return jsonify(s)


@app.route('/async/size')
def async_size():
    return jsonify({'async_size': len(app.async_workflows)})


@app.route('/async/list')
def async_list():
    return jsonify(app.async_workflows)


@app.route('/async/qsize')
def async_qsize():
    return jsonify(dict(async_qsize=async_queue.qsize()))


def get_trace_list(since=None):
    r = []
    for d in glob.glob(os.path.join(tempfile.gettempdir(), "nb2w-*")):
        if since is not None and time.time() - os.stat(d).st_mtime > since:
            continue

        try:
            summary = yaml.load(
                open(os.path.join(d, "summary.yaml")), Loader=yaml.Loader)
        except Exception as e:
            summary = "unable to load: "+repr(e)

        r.append(
            dict(
                fn=d,
                mtime=os.stat(d).st_mtime,
                ctime=os.stat(d).st_ctime,
                summary=summary,
            )
        )

    return sorted(r, key=lambda x: ['ctime'])


@app.route('/trace/size')
def trace_size():
    return jsonify({"trace_size": len(glob.glob(os.path.join(tempfile.gettempdir(), "nb2w-*")))})


@app.route('/trace/list')
def trace_list():
    f = request.args.get('format', 'json')

    if f == 'html':
        return "not implemented"
    else:
        return jsonify(get_trace_list())


@app.route('/trace/<string:job>')
def trace_get(job):
    r = []

    r = []
    for fn in glob.glob(os.path.join(tempfile.gettempdir(), job, "*_output.ipynb")):
        r.append(fn)

    return jsonify(r)


@app.route('/trace/<string:job>/<string:func>')
def trace_get_func(job, func):
    if func == "custom.css":
        return ""

    from nbconvert.exporters import HTMLExporter
    exporter = HTMLExporter()

    fn = os.path.join(tempfile.gettempdir(), job, func+"_output.ipynb")

    output, resources = exporter.from_filename(fn)

    return output


@app.route('/clear-cache')
def clear_cache():
    n_entries = None
    try:
        n_entries = len(cache.cache._cache)
    except Exception as e:
        pass

    cache.clear()

    if n_entries is not None:
        return 'cleared %i entries' % n_entries
    else:
        return 'cleared some entries'


if __name__ == '__main__':
    main()



import pytest
import nb2workflow

@pytest.fixture
def app():
    app = nb2workflow.service.app
    app.notebook_adapters = nb2workflow.nbadapter.find_notebooks('tests/testfiles')
    nb2workflow.service.setup_routes(app)
    print("creating app")
    return app

def test_boolean_default(client):
    r = client.get('/api/v1.0/get/testbool')
    assert r.json['output']['output'] == 'boolean True'
    
def test_boolean_false(client):
    r = client.get('/api/v1.0/get/testbool', query_string={'boolpar': 'false'})
    assert r.json['output']['output'] == 'boolean False'
    

def test_boolean_False(client):
    r = client.get('/api/v1.0/get/testbool', query_string={'boolpar': 'False'})
    assert r.json['output']['output'] == 'boolean False'
    
def test_boolean_empty(client):
    r = client.get('/api/v1.0/get/testbool', query_string={'boolpar': ''})
    assert r.json['output']['output'] == 'boolean False'
    
def test_boolean_true(client):
    r = client.get('/api/v1.0/get/testbool', query_string={'boolpar': 'true'})
    assert r.json['output']['output'] == 'boolean True'
    
def test_boolean_wrong(client):
    r = client.get('/api/v1.0/get/testbool', query_string={'boolpar': 'spam'})
    assert r.json['issues'][0] == 'Parameter boolpar value "spam" can not be interpreted as boolean.'
    


import os
import pytest
import base64
import os
import logging

#logger=logging.getLogger("nb2workflow")

FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)
logger=logging.getLogger("nb2workflow")
logger.setLevel(level=logging.DEBUG)


def test_nb(test_notebook):
    import nb2workflow.nbadapter
    
    basename = os.path.basename(test_notebook).replace(".ipynb","_output")

    print("basename:", basename)

    for p in ".ipynb", ".json", ".html":
        print("base and prefix", basename+p)
        if os.path.exists(basename+p):
            os.remove(basename+p)

    r = nb2workflow.nbadapter.nbrun(test_notebook, {})



    for p in ".ipynb", ".json", ".html":
        assert os.path.exists(basename+p)

    assert 'output_notebook' not in r
    assert 'output_notebook_content' not in r
    assert 'output_notebook_html' in r
    assert 'output_notebook_html_content' in r
    

def test_nb_fail(test_notebook):
    import nb2workflow.nbadapter

    try:
        nb2workflow.nbadapter.nbrun(test_notebook, dict(emin=-1))
    except Exception as e:
        print("got", e)
        assert os.path.exists(os.path.dirname(test_notebook).replace(".ipynb","_output.ipynb"))
        assert os.path.exists(os.path.dirname(test_notebook).replace(".ipynb","_exceptions.json"))
    else:
        raise Exception("failing notebook did not raise exception!")

    



from setuptools import setup
import sys

setup_requires = ['setuptools >= 30.3.0', 'setuptools-git-version']
if {'pytest', 'test', 'ptr'}.intersection(sys.argv):
    setup_requires.append('pytest-runner')

setup(name='nb2workflow',
      version = '1.3.22',
      description='convert notebook to workflow',
      author='Volodymyr Savchenko',
      author_email='contact@volodymyrsavchenko.com',
      license='GPLv3',
      packages=['nb2workflow'],
      zip_safe=False,

      entry_points={
          'console_scripts': [
            'nb2service=nb2workflow.service:main',
            'nb2worker=nb2workflow.container:main',
            'nb2workflow-version=nb2workflow:version',
            'nb2cwl=nb2workflow.cwl:main',
            'nb2rdf=nb2workflow.ontology:main',
            'nb2deploy=nb2workflow.deploy:main',
            'nbrun=nb2workflow.nbadapter:main',
            'nbinspect=nb2workflow.nbadapter:main_inspect',
            'nbreduce=nb2workflow.nbadapter:main_reduce',
            ]
      },
      
      extras_require={
        "service":[
            'flask',
            'pytest-flask',
            'flask-caching', #'Flask-Caching',
            'flask-cors',
            'flasgger',
            'python-consul',
            'apscheduler',
        ],
        "rdf":[
            'rdflib',
            'owlready2==0.11',
            'oda-knowledge-base',
        ],
        "cwl":[
            "cwlgen",
        ],
        "docker":[
            'docker',
            'checksumdir',
        ],
        "domains":[
            'numpy',
            'astropy',
            'pandas',
        ],
      },

      tests_require=[
        'pytest',
        'pytest-xprocess',
        'matplotlib',
        'cwl-runner',
      ],

      install_requires=[
        'papermill',
        'ipykernel',
        'nbconvert', 
        'psutil',
        'diskcache',
        'requests',
        'pyyaml',
        'nteract-scrapbook', 
        'werkzeug==2.0.3'
      ],


      url = 'https://github.com/volodymyrss/nb2workflow',
      download_url = 'https://github.com/volodymyrss/nb2workflow/archive/1.0.1.tar.gz',
      keywords = ['jupyter', 'docker'],
      classifiers = [],
      setup_requires=setup_requires)






from __future__ import print_function
import os
import logging
import os
import pytest
import base64
from importlib import reload

from nb2workflow import service
import nb2workflow.nbadapter
from flask import url_for

FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)
logger=logging.getLogger("nb2workflow")
logger.setLevel(level=logging.DEBUG)

def test_workflow_localfile(test_notebook_repo):
    
    from nb2workflow import workflows

    result = workflows.evaluate("localfile", test_notebook_repo, "workflow-notebook")

    print(result)
    assert result['output']
    assert len(result['output']) == 4
    assert result['exceptions'] == []

    assert 'spectrum' in result['output']

@pytest.mark.xfail
def test_workflow_exception_localfile(test_notebook):
    
    from nb2workflow import workflows

    result = workflows.evaluate("localfile", test_notebook, "workflow-notebook", scwid="66500220010.001")

    print(result)
    assert len(result['output']) == 0
    assert len(result['exceptions']) == 1

    ex = result['exceptions'][0]

    print(ex)

    

@pytest.mark.service
def test_service(service_fixture, app):
    from nb2workflow import workflows

    result = workflows.evaluate("host", service_fixture, "workflow-notebook")

    print(result)
    assert result['output']
    assert len(result['output']) == 4
    assert result['exceptions'] == []

    assert 'spectrum' in result['output']

@pytest.mark.service
def test_workflow_exception_service(service_fixture, app):
    from nb2workflow import workflows

    result = workflows.evaluate("host", service_fixture, "workflow-notebook", scwid="66500220010.001")

    print(result)
    assert len(result['output']) == 0
    assert len(result['exceptions']) == 1

    ex = result['exceptions'][0]

    print(ex)

@pytest.mark.service
def test_async_service(service_fixture, app):
    from nb2workflow import workflows

    result = workflows.evaluate("host", service_fixture, "workflow-notebook", _async_request = True)

    print(result)
    assert result['output']
    assert len(result['output']) == 4
    assert result['exceptions'] == []

    assert 'spectrum' in result['output']

@pytest.mark.service
def test_async_service_exception(service_fixture, app):
    from nb2workflow import workflows

    result = workflows.evaluate("host", service_fixture, "workflow-notebook", _async_request = True, scwid="66500220010.001")

    print(result)
    assert len(result['output']) == 0
    assert len(result['exceptions']) == 1

    ex = result['exceptions'][0]

    print(ex)



import os
import json
import logging
import pytest
import tempfile


# this can be also set in pytest call
FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger("nb2workflow")
logger.setLevel(level=logging.DEBUG)

def _mimick_convert_minor_version(nb_fn, version):
    j = json.load(open(nb_fn))
    j['nbformat_minor'] = version

    fn = nb_fn.replace(".ipynb", f"_mock_downgraded_to_{version}.ipynb")

    json.dump(j, open(fn, "w"))

    return fn

@pytest.mark.parametrize("morph_notebook", ["mimick_convert_minor_version", "vanilla"])
def test_nbadapter(test_notebook, morph_notebook, caplog):
    from nb2workflow.nbadapter import NotebookAdapter

    if morph_notebook == "mimick_convert_minor_version":
        fn = _mimick_convert_minor_version(test_notebook, 2)
    elif morph_notebook == "vanilla":
        fn = test_notebook
    else:
        raise NotImplementedError


    nba = NotebookAdapter(fn)
    parameters = nba.extract_parameters()

    for k, v in parameters.items():
        print("\033[31m", k, ":", v, "\033[0m")

    assert len(parameters) == 5

    assert 'comment' in parameters['scwid']
    assert parameters['scwid']['owl_type'] == "http://odahub.io/ontology/integral#ScWID"

    assert parameters['enabled']['owl_type'] == "http://www.w3.org/2001/XMLSchema#bool"

    outputs = nba.extract_output_declarations()
    print("outputs", outputs)

    assert len(outputs) == 3

    if os.path.exists(nba.output_notebook_fn):
        os.remove(nba.output_notebook_fn)

    if os.path.exists(nba.preproc_notebook_fn):
        os.remove(nba.preproc_notebook_fn)

    nba.execute(dict())

    output = nba.extract_output()

    if morph_notebook == "mimick_convert_minor_version":
        assert 'will attempt to convert, but expect other warnings' in caplog.text

        os.remove(fn)

    print(output)
    assert len(output) == 4

    assert 'spectrum' in output


def test_nbadapter_repo(test_notebook_repo):
    from nb2workflow.nbadapter import NotebookAdapter, find_notebooks

    nbas = find_notebooks(test_notebook_repo)

    assert len(nbas) == 1

    for nba_name, nba in nbas.items():
        print("notebook", nba_name)

        parameters = nba.extract_parameters()

        print(parameters)
        assert len(parameters) == 5

        if os.path.exists(nba.output_notebook_fn):
            os.remove(nba.output_notebook_fn)

        if os.path.exists(nba.preproc_notebook_fn):
            os.remove(nba.preproc_notebook_fn)

        nba.execute(dict())

        output = nba.extract_output()

        print(output)

        assert len(output) == 4
        assert 'spectrum' in output


def test_nbreduce(test_notebook):
    from nb2workflow.nbadapter import NotebookAdapter, nbreduce, setup_logging

    setup_logging()

    nba = NotebookAdapter(test_notebook)

    if os.path.exists(nba.output_notebook_fn):
        os.remove(nba.output_notebook_fn)

    if os.path.exists(nba.preproc_notebook_fn):
        os.remove(nba.preproc_notebook_fn)

    nba.execute(dict())

    output = nba.extract_output()

    assert len(output) == 4

    assert 'spectrum' in output

    print("will reduce", nba.output_notebook_fn)

    nbsize_b = os.path.getsize(nba.output_notebook_fn)

    assert isinstance(nbsize_b, int)

    assert nbsize_b > 0

    nbreduce(nba.output_notebook_fn, nbsize_b/1024./1024 + 1.)

    nbreduce(nba.output_notebook_fn, nbsize_b/1024./1024*0.5)


def test_denumpyfy():
    import numpy as np
    import json
    from nb2workflow.nbadapter import denumpyfy

    data = {"d": np.bool_(True), "k": np.array([1, 2, 3]), "dd": [
        1, 2, np.float32(33), {'a': np.int64(10)}]}

    try:
        r = json.dumps(data)
        print(r)
    except TypeError as e:
        print("failed as expected", e)
    else:
        raise Exception("did not fail")

    r = json.dumps(denumpyfy(data))
    print(r)



from __future__ import print_function

import os
import json
import glob
import time
import logging
import inspect
import requests
import base64
import hashlib
import datetime
import tempfile
import nbformat


logger=logging.getLogger('nb2workflow.health')

def current_health():
    issues=[]
    status = {}

    statvfs = os.statvfs(".")
    status['fs_space'] = dict(
        size_mb = statvfs.f_frsize * statvfs.f_blocks / 1024 / 1024,
        avail_mb = statvfs.f_frsize * statvfs.f_bavail / 1024 / 1024,
    )

    if status['fs_space']['avail_mb'] < 300:
        issues.append("not enough free space: %.5lg Mb left"%status['fs_space']['avail_mb'])


    import psutil

    processes = []
    status['n_open_files'] = 0
    status['n_processes'] = 0
    status['n_threads'] = 0

    for proc in psutil.process_iter():
        try:
            processes.append(dict(
                    n_open_files = len(proc.open_files()),
                ))

            status['n_open_files'] += len(proc.open_files())
            status['n_processes'] += 1
            status['n_threads'] += proc.num_threads()
        except Exception as e:
            pass

    status['cpu_times'] = dict(psutil.cpu_times_percent()._asdict())

    status['loadavg'] = psutil.getloadavg()

    if max(status['loadavg']) > 10:
        issues.append("high load avg: %s"%repr(status['loadavg']))


    status['disk_usage'] = dict([ (k+"_mb", v/1024/1024) if k!="percent" else (k,v) for k,v in dict(psutil.disk_usage(".")._asdict()).items()])

    #status['processes'] = processes

    return status, issues




import os
import logging
import pytest

test_notebook_repo=os.environ.get('TEST_NOTEBOOK_REPO')

#logger=logging.getLogger("nb2workflow")

FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)
logger=logging.getLogger("nb2workflow")
logger.setLevel(level=logging.DEBUG)

@pytest.mark.skipif(os.environ.get("TRAVIS", 'false') == "true", reason="no way of currently testing this")
def test_nbadapter_repo(test_notebook_repo):
    from nb2workflow.nbadapter import NotebookAdapter, find_notebooks
    from nb2workflow import ontology

    nbas=find_notebooks(test_notebook_repo)

    logger.info(ontology.service_semantic_signature(nbas))

        
